@@ -17,6 +17,11 @@ import six
 import pkg_resources
 import requests
 
+try:
+    from bs4 import BeautifulSoup
+except ImportError:
+    BeautifulSoup = None
+
 try:
     import argcomplete
 except ImportError:
@@ -28,7 +33,10 @@ from copr.v3 import (
     CoprNoConfigException, CoprConfigException, CoprNoResultException,
 )
 from copr.v3.pagination import next_page
-from .util import ProgressBar, json_dumps, serializable
+from copr_cli.helpers import cli_use_output_format
+from copr_cli.monitor import cli_monitor_parser
+from copr_cli.printers import cli_get_output_printer as get_printer
+from .util import ProgressBar, serializable
 from .build_config import MockProfile
 
 
@@ -73,13 +81,6 @@ Hint: {1}
 
 """
 
-output_format_help = """
-Set the formatting style. We recommend using json, which prints the required data
-in json format. The text format prints the required data in a column, one piece of
-information per line. The text-row format prints all information separated
-by a space on a single line.
-"""
-
 try:
     input = raw_input
 except NameError:
@@ -131,108 +132,6 @@ def buildopts_from_args(args, progress_callback):
     return buildopts
 
 
-class AbstractPrinter(object):
-    """Abstract class defining mandatory methods of printer classes"""
-
-    def __init__(self, fields):
-        """Represents the data we want to print.
-        Supports callable lambda function which takes exactly one argument"""
-        self.fields = fields
-
-    def add_data(self, data):
-        """Initialize the data to be printed"""
-        raise NotImplementedError
-
-    def finish(self):
-        """Print the data according to the set format"""
-        raise NotImplementedError
-
-
-class RowTextPrinter(AbstractPrinter):
-    """The class takes care of printing the data in row text format"""
-
-    def finish(self):
-        pass
-
-    def add_data(self, data):
-        row_data = []
-        for field in self.fields:
-            if callable(field):
-                row_data.append(str(field(data)))
-            else:
-                row_data.append(str(data[field]))
-        print("\t".join(row_data))
-
-
-class ColumnTextPrinter(AbstractPrinter):
-    """The class takes care of printing the data in column text format"""
-
-    first_line = True
-
-    def finish(self):
-        pass
-
-    def add_data(self, data):
-        if not self.first_line:
-            print()
-        self.first_line = False
-        for field in self.fields:
-            if callable(field):
-                print("{0}: {1}".format(field.__code__.co_varnames[0], str(field(data))))
-            else:
-                print("{0}: {1}".format(field, str(data[field])))
-
-
-class JsonPrinter(AbstractPrinter):
-    """The class takes care of printing the data in json format"""
-
-    def __init__(self, fields):
-        super(JsonPrinter, self).__init__(fields)
-        self.output_info = []
-        self.json_values_list = []
-
-    def __process_fields(self, data):
-        desired_info = []
-        for field in self.fields:
-            if callable(field):
-                desired_info.append(field(data))
-            else:
-                desired_info.append(data[field])
-        self.output_info.append(desired_info)
-
-    def add_data(self, data):
-        json_values = {}
-        self.__process_fields(data)
-        for item in self.output_info:
-            for field, info in zip(self.fields, item):
-                if callable(field):
-                    json_values[field.__code__.co_varnames[0]] = info
-                else:
-                    json_values[field] = info
-        self.json_values_list.append(json_values)
-
-    def finish(self):
-        print(json_dumps(self.json_values_list[0]))
-
-
-class JsonPrinterListCommand(JsonPrinter):
-    """The class takes care of printing the data in list in json format"""
-
-    def finish(self):
-        print(json_dumps(self.json_values_list))
-
-
-def get_printer(output_format, fields, list_command=False):
-    """According to output_format decide which object of printer to return"""
-    if output_format == "json":
-        if list_command:
-            return JsonPrinterListCommand(fields)
-        return JsonPrinter(fields)
-    if output_format == "text":
-        return ColumnTextPrinter(fields)
-    return RowTextPrinter(fields)
-
-
 class Commands(object):
     def __init__(self, config_path):
         self.config_path = config_path or '~/.config/copr'
@@ -292,6 +191,14 @@ class Commands(object):
             owner = self.config["username"]
         return owner, name
 
+    def build_url(self, build_id):
+        """
+        Return the "generic" predictable url for build_id, which redirects
+        to the final owner/project/build_id route.
+        """
+        return urljoin(self.config["copr_url"],
+                       "/coprs/build/{0}".format(build_id))
+
     def parse_chroot_path(self, path):
         """
         Take a `path` in an `owner/project/chroot` format and return a tuple of
@@ -311,6 +218,7 @@ class Commands(object):
 
         prevstatus = defaultdict(lambda: None)
         failed_ids = []
+        canceled_ids = []
 
         watched = set(build_ids)
         done = set()
@@ -334,6 +242,8 @@ class Commands(object):
 
                     if build_details.state in ["failed"]:
                         failed_ids.append(build_id)
+                    if build_details.state in ["canceled"]:
+                        canceled_ids.append(build_id)
                     if build_details.state in ["succeeded", "skipped",
                                                "failed", "canceled"]:
                         done.add(build_id)
@@ -346,10 +256,16 @@ class Commands(object):
 
                 time.sleep(30)
 
+            exception_message = ""
+            separator = ""
             if failed_ids:
-                raise copr_exceptions.CoprBuildException(
-                    "Build(s) {0} failed.".format(
-                        ", ".join(str(x) for x in failed_ids)))
+                exception_message = "Build(s) {0} failed.".format(", ".join(str(x) for x in failed_ids))
+                separator = " "
+            if canceled_ids:
+                exception_message += separator + "Build(s) {0} canceled.".format(", ".join(str(x) for x in canceled_ids))
+
+            if failed_ids or canceled_ids:
+                raise copr_exceptions.CoprBuildException(exception_message)
 
         except KeyboardInterrupt:
             pass
@@ -504,8 +420,7 @@ class Commands(object):
             print("Build was added to {0}:".format(builds[0].projectname))
 
             for build in builds:
-                url = urljoin(self.config["copr_url"], "/coprs/build/{0}".format(build.id))
-                print("  {0}".format(url))
+                print("  {0}".format(self.build_url(build.id)))
 
             build_ids = [build.id for build in builds]
             print("Created builds: {0}".format(" ".join(map(str, build_ids))))
@@ -610,6 +525,20 @@ class Commands(object):
             print("Updating packages in {0} from {1}/{2}.\nPlease be aware that it may take a few minutes "
                   "to duplicate backend data.".format(project.full_name, srcownername, srcprojectname))
 
+    @requires_api_auth
+    def action_regenerate_repos(self, args):
+        """ Method called when the 'regenerate-repos' action has been selected
+        by the user.
+
+        :param args: argparse arguments provided by the user
+        """
+        ownername, projectname = self.parse_name(args.copr)
+        project = self.client.project_proxy.regenerate_repos(
+            ownername, projectname)
+
+        print("Submitted action to regenerate repository metadata for "
+              "the '{0}' project.".format(project.full_name))
+
     def action_list_builds(self, args):
         """ Method called when the 'list-builds' action has been selected by
         the user.
@@ -617,7 +546,7 @@ class Commands(object):
         :param args: argparse arguments provided by the user
         """
         ownername, projectname = self.parse_name(args.project)
-        pagination = {"limit": 100}
+        pagination = {"limit": 1000}
 
         builds_list = self.client.build_proxy.get_list(ownername, projectname,
                                                        pagination=pagination)
@@ -626,13 +555,13 @@ class Commands(object):
             sys.stderr.write(
                 "The default setting will be changed from text-row to json in the following releases\n")
 
+        fields = ["id", lambda name: name["source_package"]["name"], "state"]
+        printer = get_printer(args.output_format, fields, True)
         while builds_list:
-            fields = ["id", lambda name: name["source_package"]["name"], "state"]
-            printer = get_printer(args.output_format, fields, True)
             for data in builds_list:
                 printer.add_data(data)
-            printer.finish()
             builds_list = next_page(builds_list)
+        printer.finish()
 
     def action_mock_config(self, args):
         """ Method called when the 'mock-config' action has been selected by the
@@ -862,17 +791,35 @@ class Commands(object):
 
     def action_list_packages(self, args):
         ownername, projectname = self.parse_name(args.copr)
-        packages = self.client.package_proxy.get_list(ownername=ownername, projectname=projectname,
-                                                      with_latest_build=args.with_latest_build,
-                                                      with_latest_succeeded_build=args.with_latest_succeeded_build)
-        packages_with_builds = [self._package_with_builds(p, args) for p in packages]
-        fields = ["id", "auto_rebuild", "name", "ownername", "projectname", "source_dict",
-                  "source_type", "latest_succeeded_build", "latest_build"]
+        fields = [
+            "id",
+            "auto_rebuild",
+            "name",
+            "ownername",
+            "projectname",
+            "source_dict",
+            "source_type",
+            "latest_succeeded_build",
+            "latest_build"
+        ]
         if args.with_all_builds:
             fields.insert(1, "builds")
+
+        pagination = {"limit": 1000}
+        packages = self.client.package_proxy.get_list(
+            ownername=ownername, projectname=projectname,
+            with_latest_build=args.with_latest_build,
+            with_latest_succeeded_build=args.with_latest_succeeded_build,
+            pagination=pagination,
+        )
+
         printer = get_printer(args.output_format, fields, True)
-        for data in packages_with_builds:
-            printer.add_data(data)
+        while packages:
+            packages_with_builds = [self._package_with_builds(p, args)
+                                    for p in packages]
+            for data in packages_with_builds:
+                printer.add_data(data)
+            packages = next_page(packages)
         printer.finish()
 
     def action_list_package_names(self, args):
@@ -1199,12 +1146,21 @@ def setup_parser():
     parser_delete.add_argument("--confirm", action="store_true", help="Confirm forking into existing project")
     parser_delete.set_defaults(func="action_fork")
 
+    # create the parser for the "regenerate-repos" command
+    parser_regenerate_repos = subparsers.add_parser(
+        "regenerate-repos",
+        help=("Regenerate repository metadata for a project. "
+              "Useful when automatic generation is disabled"))
+    parser_regenerate_repos.add_argument(
+        "copr",
+        help=("Can be just name of project or even in format owner/project."))
+    parser_regenerate_repos.set_defaults(func="action_regenerate_repos")
+
     parser_builds = subparsers.add_parser("list-builds", help="List all builds in the project")
     parser_builds.add_argument("project", help="Which project's builds should be listed.\
                                Can be just a name of the project or even in format\
                                username/project or @groupname/project.")
-    parser_builds.add_argument("--output-format", choices=["text", "json", "text-row"],
-                               help=output_format_help)
+    cli_use_output_format(parser_builds, default=None)
     parser_builds.set_defaults(func="action_list_builds")
 
     #########################################################
@@ -1470,8 +1426,7 @@ def setup_parser():
 
     parser_get_chroot = subparsers.add_parser("get-chroot", help="Get chroot of a project")
     parser_get_chroot.add_argument("coprchroot", help="Path to a project chroot as owner/project/chroot or project/chroot")
-    parser_get_chroot.add_argument("--output-format", default="json", choices=["text", "json", "text-row"],
-                                   help=output_format_help)
+    cli_use_output_format(parser_get_chroot)
     parser_get_chroot.set_defaults(func="action_get_chroot")
 
     parser_list_chroots = subparsers.add_parser("list-chroots", help="List all currently available chroots.")
@@ -1595,8 +1550,7 @@ def setup_parser():
                                       help="Also display data related to the latest succeeded build for the package.")
     parser_list_packages.add_argument("--with-all-builds", action="store_true",
                                       help="Also display data related to the builds for the package.")
-    parser_list_packages.add_argument("--output-format", default="json", choices=["text", "json", "text-row"],
-                                      help=output_format_help)
+    cli_use_output_format(parser_list_packages)
     parser_list_packages.set_defaults(func="action_list_packages")
 
     # package names listing
@@ -1620,8 +1574,7 @@ def setup_parser():
                                     help="Also display data related to the latest succeeded build for each package.")
     parser_get_package.add_argument("--with-all-builds", action="store_true",
                                     help="Also display data related to the builds for each package.")
-    parser_get_package.add_argument("--output-format", default="json", choices=["text", "json", "text-row"],
-                                    help=output_format_help)
+    cli_use_output_format(parser_get_package)
     parser_get_package.set_defaults(func="action_get_package")
 
     # package deletion
@@ -1751,6 +1704,9 @@ def setup_parser():
             help=request_help.format('builder'))
     parser_permissions_request.set_defaults(func='action_permissions_request')
 
+    # package monitoring
+    cli_monitor_parser(subparsers)
+
     if argcomplete:
         argcomplete.autocomplete(parser)
     return parser
@@ -1774,6 +1730,27 @@ def str2bool(v):
         raise argparse.ArgumentTypeError("Boolean value expected.")
 
 
+def _handle_frontend_api_request_error(e, args):
+    if "Request is not in JSON format" not in str(e):
+        return
+    if not args.debug:
+        sys.stderr.write("Try 'copr-cli --debug' for more info.\n")
+        return
+
+    page_content = e.result.__response__.content
+    if BeautifulSoup:
+        soup = BeautifulSoup(page_content, features="html.parser")
+        page_content = soup.get_text()
+    else:
+        page_content = re.sub(r'<.*?>', '', page_content.decode("utf-8"))
+
+    sys.stderr.write(
+        "\n"
+        "Server response:\n"
+        "----------------\n"
+        "%s\n" % page_content)
+
+
 def main(argv=sys.argv[1:]):
     try:
         # Set up parser for global args
@@ -1788,7 +1765,12 @@ def main(argv=sys.argv[1:]):
             return
 
         commands = Commands(arg.config)
-        getattr(commands, arg.func)(arg)
+        if isinstance(arg.func, str):
+            # Call self method by its name
+            getattr(commands, arg.func)(arg)
+        else:
+            # Call external command method
+            arg.func(commands, arg)
 
     except KeyboardInterrupt:
         sys.stderr.write("\nInterrupted by user.")
@@ -1802,6 +1784,7 @@ def main(argv=sys.argv[1:]):
     except (CoprRequestException, CoprNoResultException) as e:
         sys.stderr.write("\nSomething went wrong:")
         sys.stderr.write("\nError: {0}\n".format(e))
+        _handle_frontend_api_request_error(e, arg)
         sys.exit(1)
     except argparse.ArgumentTypeError as e:
         sys.stderr.write("\nError: {0}\n".format(e))
